from typing import Any

def get_properties_dict(instance) -> dict[str: Any]:
    """Provide a class instance and receive its property (not attribute) names & values.
    Useful because __dict__ only returns attribute names & values"""
    cls = type(instance)
    properties = {name: getattr(instance, name) for name, value in cls.__dict__.items() if isinstance(value, property)}
    return properties

def get_attrs_and_props(instance) -> dict[str: Any]:
    return instance.__dict__ | get_properties_dict(instance)


import bs4

def list_of_dicts_from_soup_table(soup_table: bs4.element.Tag) -> list[dict]:
    """Accepts a BS table (a list of values for every row, including the header).
    Converts & returns a list of dicts where the headers are keys & the row data are values."""
    table_of_dicts = []
    rows = soup_table.find_all("tr")
    # Iterate over rows and extract cells (td or th tags), extracting the data from each cell in the row
    table_data = [[cell.get_text(strip=True) for cell in row.find_all(["td", "th"])] for row in rows]
    headers = [col for col in table_data[0]]
    for idx, row in enumerate(table_data[1:]):
        table_of_dicts.append({header: r for header, r in zip(headers, row)})
    return table_of_dicts



from datetime import date, datetime
from collections import Counter
from operator import itemgetter
from typing import Any, Sequence

import pandas as pd

def add_running_total_item(data: list[dict], sorts: list[str | tuple[str, bool]], value_key: str) -> list[dict]:
    """For a list of dictionaries, return a list of dicts with an added key called running_total
    whose value is the summed value_key.
    Sorts is a list whose members can be either a string e.g. 'my_col' or a tuple e.g. ('my_col', True).
    The boolean True will return that key in the reverse/descending order.
    The value for the provided value_key must be an integer or float."""
    data = sort_data(data, sorts)
    running_total = 0
    for d in data:
        for k, v in d.items():
            if k == value_key:
                d.setdefault(k, 0)
                running_total += v
                d['running_total'] = running_total
                break
    return data

def filter_data(incoming_data: list[dict], filters: dict) -> list[dict]:
    """Filters is a dict whose values are a sequence. If a value is a tuple of two dates/date-times, then the value
    must fall between the two dates/date-times. If a value is ['All', 'all', 'ALL'], it is not filtered."""
    all_terms = ['All', 'all', 'ALL']

    def is_special_date_parameter(seq: Sequence) -> bool:
        return all([isinstance(seq, tuple), len(seq) == 2, all(isinstance(elem, (date, datetime)) for elem in seq)])

    # TODO: evaluate if i should backport this to disc golf project (if it works)
    if not filters:
        return incoming_data
    filtered = incoming_data
    for key, value in filters.items():
        if value and value not in all_terms:
            if is_special_date_parameter(value):
                filtered = [entry for entry in filtered if value[0] <= entry[key] <= value[1]]
            else:
                filtered = [entry for entry in filtered if entry[key] in value]
    return filtered


def group_data(data: list[dict], grouper_keys: list[str], value_key: str) -> list[dict]:
    """ Grouper_keys is a list of keys, so multi-key grouping is supported.
    The value_key's value must be an int or float. """
    # TODO: backport to Disc Golf project, if this works
    counter = Counter()

    def is_summable(val):
        return isinstance(val, (int, float))

    for item in data:
        if not is_summable(item[value_key]):
            raise ValueError('The value of the value_col must be an integer or float')

        key = tuple(item[grouper] for grouper in grouper_keys)
        value = item[value_key]
        counter[key] += value

    result = []
    for key, total in counter.items():
        group_dict = dict(zip(grouper_keys, key))
        group_dict[value_key] = total
        result.append(group_dict)

    return result

def sort_data(data: list[dict], sorts: list[str | tuple[str, bool]]) -> list[dict]:
    """Sorts is a list whose members can be either a string e.g. 'my_col' or a tuple e.g. ('my_col', True).
    The boolean True will return that key in the reverse/descending order"""
    # for each sort item, if the value is a string (the column name), convert it to a tuple (col_name, True))
    sort_tuples = [s if isinstance(s, tuple) else (s, False) for s in sorts]
    # iterate backwards through the sort tuples & perform each sort
    for key, reverse in reversed(sort_tuples):
        data.sort(key=itemgetter(key), reverse=reverse)
    return data

def to_df(data: list[dict[str, Any]]) -> pd.DataFrame:
    return pd.DataFrame(data)

def unique_sorted(values: list[dict], key: str) -> list[Any]:
    """From a list of dicts, return a list of unique sorted values for a given key.
    The key doesn't have to be present in any or all of the dicts."""
    return sorted({v for d in values for k, v in d.items() if k == key and v})



import html
import re

import requests


def scrape_pages_until_text_not_found(base_url: str, required_text: str,
                                      begin_page_num: int = 0, end_page_num: int = 100) -> list[str]:
    """Ex base_url: http://www.example.com/page/  ex required_text: 'item_price' ...
    for range(begin_page_num, end_page_num), if the page contains 'item_price' ...
    return each html as a string in a list of pages"""
    page_scrapes: list[str] = []
    for i in range(begin_page_num, end_page_num + 1):
        url_w_page_number = f'{base_url}{i}'
        print('Scraping', url_w_page_number)
        page_scrape = requests.get(url_w_page_number).text
        print(f'Found {len(page_scrape)} characters')
        if required_text not in page_scrape:
            print(f'Stopping scrape as required text: "{required_text}" not found on page: {url_w_page_number}')
            break
        page_scrapes.append(page_scrape)
    return page_scrapes


def find_all_matched(html_str: str, preceding_text: list[str] | str, succeeding_text: str) -> list[str]:
    """string of 'abcdefg xZfg', preceding_texts = ['ab', 'x'] & succeeding_text = 'fg', returns ['cde', 'Z'].
    succeeding text should be r-string"""
    # Regex pattern: look for any of the targets and capture everything until next quote
    if isinstance(preceding_text, list):
        pattern = re.compile(r'(?:' + '|'.join(map(re.escape, preceding_text)) + f')(.*?){succeeding_text}')
    else:
        pattern = re.compile(re.escape(preceding_text) + f'(.*?){succeeding_text}')
    return pattern.findall(html_str)

def find_all_matched_until(html_str: str, preceding_text: list[str] | str, succeeding_text: str, until_text: str) -> list[str]:
    """like find_all_matched() except it will stop looking when it encounters until_text"""
    until_text_pos: int = html_str.find(until_text)
    if until_text_pos == -1:
        return find_all_matched(html_str, preceding_text, succeeding_text)
    html_str = html_str[:until_text_pos]
    return find_all_matched(html_str, preceding_text, succeeding_text)


def find_first_match(html_str: str, preceding_text: list[str] | str, succeeding_text: str) -> str | None:
    """ex html_str='abcdefg xZfg', preceding_text='abc', succeeding_text='Zfg', returns 'defg x' """
    if isinstance(preceding_text, list):
        pattern = re.compile(r'(?:' + '|'.join(map(re.escape, preceding_text)) + f')(.*?){succeeding_text}')
    else:
        pattern = re.compile(re.escape(preceding_text) + f'(.*?){succeeding_text}')
    match = re.search(pattern, html_str)
    return match.group(1) if match else None


def clean_raw_html(text: str) -> str:
    """handles HTML entities like &amp;, &#x27; replaces characters like \\\\\n, \\t, \\\\ with a space"""
    tmp1 = html.unescape(text)  # html.unescape handles HTML entities like &amp;, &#x27;
    tmp2 = re.sub(r'\\+n', ' ', tmp1)  # handles escaped characters like \\\\n, \\t, \\\"
    return tmp2



from typing import Any

import pandas as pd

def to_df(data: list[dict[str, Any]]) -> pd.DataFrame:
    return pd.DataFrame(data)

def color_value_map(val: Any, colors: tuple[str, str, str] = ('green', 'red', 'white')) -> str:
    """Use with df.style.map.  For any column with a value of int or float, return the colored text string.
    Example usage: colored_df = df.style.map(color_value_map)"""
    positive, negative, neutral = colors
    if isinstance(val, (int, float)):
        color = negative if val < 0 else positive if val > 0 else neutral
        return f'color: {color}'



def list_items_in_english(items: list) -> str | None:
    """[] -> None, ['Alice'] -> 'Alice', ['Alice', 'Bob'] -> 'Alice and Bob',
    ['Alice', 'Bob', 'Charlie'] -> 'Alice, Bob, and Charlie'
    ['Alice', 'Bob', 'Charlie', 'David'] -> 'Alice, Bob, and 2 others'"""
    item_cnt = len(items)
    if item_cnt == 0:
        return None
    elif item_cnt == 1:
        return f'{items[0]}'
    elif item_cnt == 2:
        return f'{items[0]} and {items[1]}'
    elif item_cnt == 3:
        return f'{items[0]}, {items[1]} and {items[2]}'
    else:
        return f'{items[0]}, {items[1]} and {len(items) - 2} others'
